//////////////////////////////////////////////////////////////////////
// SimpleRobot.cpp: implementation for the Car class (Hierarchical Transformation).
// -Kyoung Shin Park (2013 Spring)
//////////////////////////////////////////////////////////////////////
#include "simpleRobot.h"

namespace moglclass
{
	
SimpleRobot::SimpleRobot()
{
	theta = 0.0f; // 0~360
	phi = 0.0f; // 0~360
	psi = 0.0f; // 0~360

	active = false;

	init();
}

SimpleRobot::~SimpleRobot()
{ 
	if (base) delete base;
	if (arm[0]) delete arm[0];
	if (arm[1]) delete arm[1];
	std::cout << "~SimpleRobot" << std::endl; 
}

void SimpleRobot::init()
{
	base = new Cylinder(glm::vec3(0, 0, 0), 2.0f, 3.0f, 16);
	base->setColor(glm::vec3(1,0,0));
	arm[0] = new Parallelepiped(glm::vec3(-0.25f, 0.0f, -0.25f), glm::vec3(0.5f, 0.0f, 0.0f), glm::vec3(0.0f, 0.0f, 0.5f), glm::vec3(0.0f, 3.0f, 0.0f));
	arm[0]->setColor(glm::vec3(0,1,0));
	arm[1] = new Parallelepiped(glm::vec3(-0.25f, 0.0f, -0.25f), glm::vec3(0.5f, 0.0f, 0.0f), glm::vec3(0.0f, 0.0f, 0.5f), glm::vec3(0.0f, 3.0f, 0.0f));
	arm[1]->setColor(glm::vec3(0,0,1));
}

void SimpleRobot::draw(Program* p, glm::mat4& projection, glm::mat4& view, glm::mat4& model)
{
	p->useProgram();
	p->setUniform("gProjection", projection);
	p->setUniform("gView", view);

	// base
	glm::mat4 baseMatrix = model * baseTransform;
	p->setUniform("gModel", baseMatrix);
	base->draw();

	// lowerArm
	glm::mat4 lowerArmMatrix = model * baseTransform * lowerArmTransform;
	p->setUniform("gModel", lowerArmMatrix);
	arm[0]->draw();

	// upperArm
	glm::mat4 upperArmMatrix  = model * baseTransform * lowerArmTransform * upperArmTransform;
	p->setUniform("gModel", upperArmMatrix);
	arm[1]->draw();
}

bool SimpleRobot::update(float deltaTime)
{
	if (active)
	{
		theta += 0.1f * deltaTime;
		phi += 0.1f * deltaTime;
		psi += 0.1f * deltaTime;
	}

	baseTransform = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, -1.5f, 0.0f)) * glm::rotate(glm::mat4(1.0f), theta, glm::vec3(0.0f, 1.0f, 0.0f));
	lowerArmTransform = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 1.5f, 0.0f)) * glm::rotate(glm::mat4(1.0f), phi, glm::vec3(0.0f, 0.0f, 1.0f)); 
	upperArmTransform = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 3.0f, 0.0f)) * glm::scale(glm::mat4(1.0f), glm::vec3(1.0f, 0.5f, 1.0f)) * glm::rotate(glm::mat4(1.0f), psi, glm::vec3(0.0f, 0.0f, 1.0f)); 

	return true;
}

} // moglclass namespace